function [x, iter, error] = bicgstabm4(PSF,aPSF,A,Ba,Bb,Bc,Bd,x,b, max_it, tol,m1,m2,pm,pn)

%  -- Iterative template routine --
%     Univ. of Tennessee and Oak Ridge National Laboratory
%     October 1, 1993
%     Details of this algorithm are described in "Templates for the
%     Solution of Linear Systems: Building Blocks for Iterative
%     Methods", Barrett, Berry, Chan, Demmel, Donato, Dongarra,
%     Eijkhout, Pozo, Romine, and van der Vorst, SIAM Publications,
%     1993. (ftp netlib2.cs.utk.edu; cd linalg; get templates.ps).
%
%  [x, error, iter, flag] = bicgstab(A, x, b, M, max_it, tol)
%
% bicgstab.m solves the linear system Ax=b using the 
% BiConjugate Gradient Stabilized Method with preconditioning.
%
% input   A        REAL matrix
%         x        REAL initial guess vector
%         b        REAL right hand side vector
%         M        REAL preconditioner matrix
%         max_it   INTEGER maximum number of iterations
%         tol      REAL error tolerance
%
% output  x        REAL solution vector
%         error    REAL error norm
%         iter     INTEGER number of iterations performed
%         flag     INTEGER: 0 = solution found to tolerance
%                           1 = no convergence given max_it
%                          -1 = breakdown: rho = 0
%                          -2 = breakdown: omega = 0

  iter = 0;                                          % initialization

  r = b - funH(x,PSF,aPSF,A,Ba,Bb,Bc,Bd,m1,m2,tol,pm,pn);
  error = norm( r,'fro'); 
  if ( error < tol ), return, end

  omega  = 1.0;
  r_tld = r;

  for iter = 1:max_it,                              % begin iteration

     rho   = r_tld(:)'*r(:) ;                          % direction vector
     if ( rho == 0.0 ), break, end

     if ( iter > 1 ),
        beta  = ( rho/rho_1 )*( gam/omega );
        p = r + beta*( p - omega*v );
     else
        p = r;
     end
 
     v = funH(p,PSF,aPSF,A,Ba,Bb,Bc,Bd,m1,m2,tol,pm,pn);
     gam = rho / (r_tld(:)'*v(:));
     s = r - gam*v;
     if ( norm(s,'fro') < tol ),                          % early convergence check
        x = x + gam*p;
        break;
     end

     t = funH(s,PSF,aPSF,A,Ba,Bb,Bc,Bd,m1,m2,tol,pm,pn);
     omega = (t(:)'*s(:)) / (t(:)'*t(:));

     x = x + gam*p + omega*s;             % update approximation

     r = s - omega*t;
     error = norm( r,'fro' ) ;                  % check convergence
     if ( error <= tol ), break, end

     if ( omega == 0.0 ), break, end
     rho_1 = rho;

  end
 
% END bicgstab.m

